#include "WMSLayer.h"

#include "../../../../dataaccess/datasource/DataSource.h"
#include "../../../../dataaccess/dataset/DataSetType.h"
#include "../../../../dataaccess/utils/Utils.h"
#include "../../../../maptools/AbstractRenderer.h"
#include "../../../../maptools/RendererFactory.h"
#include "../../../../raster/RasterProperty.h"
#include "../../../../raster/BandProperty.h"
#include "../../../../raster/Band.h"
#include "../../../../raster/Grid.h"
#include "../../../../raster/RasterFactory.h"
#include "../../../../memory/DataSet.h"
#include "../../../../memory/DataSetItem.h"
#include "../../../core/Exception.h"
#include "DataSource.h"

//Boost
#include <boost/format.hpp>

//STL
#include <vector>

const std::string te::ws::ogc::wms::WMSLayer::sm_type("OGCWMSLAYER");

te::ws::ogc::wms::WMSLayer::WMSLayer(te::map::AbstractLayer *parent)
  : te::map::AbstractLayer(parent),
    m_rendererType("OGC_WMS_LAYER_RENDERER"),
    m_currentSRID(TE_UNKNOWN_SRS)
{

}

te::ws::ogc::wms::WMSLayer::WMSLayer(const std::string &id, te::map::AbstractLayer *parent)
  : te::map::AbstractLayer(id, parent),
    m_rendererType("OGC_WMS_LAYER_RENDERER"),
    m_currentSRID(TE_UNKNOWN_SRS)
{

}

te::ws::ogc::wms::WMSLayer::WMSLayer(const std::string &id, const std::string &title, te::map::AbstractLayer *parent)
  : te::map::AbstractLayer(id, title, parent),
    m_rendererType("OGC_WMS_LAYER_RENDERER"),
    m_currentSRID(TE_UNKNOWN_SRS)
{

}

te::ws::ogc::wms::WMSLayer::~WMSLayer()
{

}

std::unique_ptr<te::da::DataSetType> te::ws::ogc::wms::WMSLayer::getDataSetType() const
{
  te::da::DataSetType* dataSetType = new te::da::DataSetType(m_datasetName);

  if(m_currentImage.get() == 0)
  {
    return std::unique_ptr<te::da::DataSetType>(dataSetType);
  }

  te::rst::RasterProperty* rp = new te::rst::RasterProperty("raster");

  dataSetType->add(rp);

  for(size_t i = 0; i < m_currentImage->getNumberOfBands(); i++)
  {
    rp->add(new te::rst::BandProperty(*m_currentImage->getBand(i)->getProperty()));
  }

  rp->set(new te::rst::Grid(*m_currentImage->getGrid()));

  return std::unique_ptr<te::da::DataSetType>(dataSetType);
}

const te::gm::Envelope &te::ws::ogc::wms::WMSLayer::getExtent() const
{
  if(m_currentSRID == TE_UNKNOWN_SRS)
  {
    return te::map::AbstractLayer::getExtent();
  }
  else
  {
    return m_currentExtent;
  }
}

int te::ws::ogc::wms::WMSLayer::getSRID() const
{
  if(m_currentSRID == TE_UNKNOWN_SRS)
  {
    return te::map::AbstractLayer::getSRID();
  }
  else
  {
    return m_currentSRID;
  }
}

void te::ws::ogc::wms::WMSLayer::setCurrentSRID(const int &srid)
{
  if(m_currentSRID == srid)
    return;

  m_currentSRID = srid;

  m_currentExtent = te::map::AbstractLayer::getExtent();

  m_currentExtent.transform(m_srid, m_currentSRID);
}

std::unique_ptr<te::map::LayerSchema> te::ws::ogc::wms::WMSLayer::getSchema() const
{
  m_schema = this->getDataSetType();

  if(m_schema.get())
    return std::unique_ptr<te::map::LayerSchema>(static_cast<te::map::LayerSchema*>(m_schema->clone()));

  assert(!m_datasetName.empty());

  te::da::DataSourcePtr ds = te::da::GetDataSource(m_datasourceId, true);

  m_schema = ds->getDataSetType(m_datasetName);

  return std::unique_ptr<te::map::LayerSchema>(static_cast<te::map::LayerSchema*>(m_schema->clone()));
}

std::unique_ptr<te::da::DataSet> te::ws::ogc::wms::WMSLayer::getData(te::common::TraverseType travType, const te::common::AccessPolicy accessPolicy) const
{
  if(m_currentImage.get() == 0)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("No data is available."));
  }

  std::unique_ptr<te::da::DataSetType> dataSetType = this->getDataSetType();

  if(dataSetType.get() == 0)
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("No data type is available."));
  }

  if(!dataSetType->hasRaster())
  {
    throw te::ws::core::Exception() << te::ErrorDescription(TE_TR("No raster data type is available."));
  }

  te::rst::RasterProperty* rp = static_cast<te::rst::RasterProperty*>(dataSetType->getProperty(0));

  std::map<std::string, std::string> rinfo;

  te::rst::Raster* memRaster = te::rst::RasterFactory::make("MEM", rp->getGrid(), rp->getBandProperties(), rinfo, m_currentImage.get());

  te::mem::DataSet* dataSet = new te::mem::DataSet(dataSetType.release());

  te::mem::DataSetItem* dataSetItem = new te::mem::DataSetItem(dataSet);

  dataSetItem->setRaster(0, memRaster);

  dataSet->add(dataSetItem);

  return std::unique_ptr<te::da::DataSet>(dataSet);
}

std::unique_ptr<te::da::DataSet> te::ws::ogc::wms::WMSLayer::getData(const std::string& /*propertyName*/,
                                                                   const te::gm::Envelope* /*e*/,
                                                                   te::gm::SpatialRelation /*r*/,
                                                                   te::common::TraverseType travType,
                                                                   const te::common::AccessPolicy accessPolicy) const
{
  return getData(travType, accessPolicy);
}

std::unique_ptr<te::da::DataSet> te::ws::ogc::wms::WMSLayer::getData(const std::string& /*propertyName*/,
                                                                   const te::gm::Geometry* /*g*/,
                                                                   te::gm::SpatialRelation /*r*/,
                                                                   te::common::TraverseType travType,
                                                                   const te::common::AccessPolicy accessPolicy) const
{
  return getData(travType, accessPolicy);
}

std::unique_ptr<te::da::DataSet> te::ws::ogc::wms::WMSLayer::getData(te::da::Expression* /*restriction*/,
                                                                   te::common::TraverseType travType,
                                                                   const te::common::AccessPolicy accessPolicy) const
{
  return getData(travType, accessPolicy);
}

std::unique_ptr<te::da::DataSet> te::ws::ogc::wms::WMSLayer::getData(const te::da::ObjectIdSet* /*oids*/,
                                                                   te::common::TraverseType travType,
                                                                   const te::common::AccessPolicy accessPolicy) const
{
  return getData(travType, accessPolicy);
}

bool te::ws::ogc::wms::WMSLayer::isValid() const
{
  if(m_datasourceId.empty() || m_getMapRequest.m_layers.empty())
    return false;

  if(m_getMapRequest.m_width == 0 || m_getMapRequest.m_height == 0 || m_getMapRequest.m_format.empty())
    return false;

  te::da::DataSourcePtr ds;
  try
  {
    ds = te::da::GetDataSource(m_datasourceId, true);
  }
  catch(...)
  {
    return false;
  }

  if(ds.get() == 0 || !ds->isValid() || !ds->isOpened())
    return false;

  return true;
}

void te::ws::ogc::wms::WMSLayer::draw(te::map::Canvas *canvas,
                                      const te::gm::Envelope &bbox, int srid,
                                      const double &scale, bool *cancel)
{
  if(m_rendererType.empty())
    throw te::ws::core::Exception() << te::ErrorDescription((boost::format(TE_TR("Could not draw the WMS layer %1%. The renderer type is empty!")) % getTitle()).str());

  // Try get the defined renderer
  std::unique_ptr<te::map::AbstractRenderer> renderer(te::map::RendererFactory::make(m_rendererType));

  if(renderer.get() == 0)
    throw te::ws::core::Exception() << te::ErrorDescription((boost::format(TE_TR("Could not draw the WMS layer %1%. The renderer %2% could not be created!"))  % getTitle() % m_rendererType).str());

  this->setCurrentSRID(srid);

  renderer->draw(this, canvas, bbox, srid, scale, cancel);
}

const std::string &te::ws::ogc::wms::WMSLayer::getType() const
{
  return sm_type;
}

const std::string &te::ws::ogc::wms::WMSLayer::getDataSourceId() const
{
  return m_datasourceId;
}

void te::ws::ogc::wms::WMSLayer::setDataSourceId(const std::string &datasourceId)
{
  m_datasourceId = datasourceId;
}

const std::string &te::ws::ogc::wms::WMSLayer::getRendererType() const
{
  return m_rendererType;
}

void te::ws::ogc::wms::WMSLayer::setRendererType(const std::string &rendererType)
{
  m_rendererType = rendererType;
}

te::ws::ogc::wms::WMSGetMapRequest te::ws::ogc::wms::WMSLayer::getRequest() const
{
  return m_getMapRequest;
}

void te::ws::ogc::wms::WMSLayer::setGetMapRequest(const te::ws::ogc::wms::WMSGetMapRequest &getMapRequest)
{
  m_getMapRequest = getMapRequest;
}

const te::ws::ogc::wms::WMSGetMapResponse te::ws::ogc::wms::WMSLayer::getMap() const
{

  te::da::DataSourcePtr ds;
  try
  {
    ds = te::da::GetDataSource(m_datasourceId, true);
  }
  catch(...)
  {
    return te::ws::ogc::wms::WMSGetMapResponse();
  }

  if(ds.get() == 0 || !ds->isValid() || !ds->isOpened())
    return te::ws::ogc::wms::WMSGetMapResponse();

  te::ws::ogc::wms::da::DataSource* wmsDataSource = dynamic_cast<te::ws::ogc::wms::da::DataSource*>(ds.get());

  if(wmsDataSource == nullptr)
  {
    return te::ws::ogc::wms::WMSGetMapResponse();
  }

  return wmsDataSource->getMap(m_getMapRequest);
}

const std::string te::ws::ogc::wms::WMSLayer::saveGetMap(const std::string &filename)
{
  te::da::DataSourcePtr ds;
  try
  {
    ds = te::da::GetDataSource(m_datasourceId, true);
  }
  catch(...)
  {
    return std::string("");
  }

  if(ds.get() == 0 || !ds->isValid() || !ds->isOpened())
    return std::string("");

  te::ws::ogc::wms::da::DataSource* wmsDataSource = dynamic_cast<te::ws::ogc::wms::da::DataSource*>(ds.get());

  if(wmsDataSource == nullptr)
  {
    return std::string("");
  }

  return wmsDataSource->saveGetMap(m_getMapRequest, filename);
}

void te::ws::ogc::wms::WMSLayer::setCurrentRaster(te::rst::Raster* getMapResponse)
{
  m_currentImage.reset(getMapResponse);
}

const te::ws::ogc::wms::WMSGetFeatureInfoResponse te::ws::ogc::wms::WMSLayer::getFeatureInfo(const te::ws::ogc::wms::WMSGetFeatureInfoRequest &request)
{
  te::da::DataSourcePtr ds;
  try
  {
    ds = te::da::GetDataSource(m_datasourceId, true);
  }
  catch(...)
  {
    return te::ws::ogc::wms::WMSGetFeatureInfoResponse();
  }

  if(ds.get() == 0 || !ds->isValid() || !ds->isOpened())
    return te::ws::ogc::wms::WMSGetFeatureInfoResponse();

  te::ws::ogc::wms::da::DataSource* wmsDataSource = dynamic_cast<te::ws::ogc::wms::da::DataSource*>(ds.get());

  if(wmsDataSource == nullptr)
  {
    return te::ws::ogc::wms::WMSGetFeatureInfoResponse();
  }

  return wmsDataSource->getFeatureInfo(request);
}







